/*
 * Copyright (C) 2016 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the specific language governing
 * permissions and limitations under the License.
 */

package com.android.systemui.plugins;

import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import android.app.Activity;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.ContextWrapper;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.pm.ApplicationInfo;
import android.content.pm.PackageManager;
import android.content.pm.PackageManager.NameNotFoundException;
import android.content.pm.ResolveInfo;
import android.content.pm.ServiceInfo;
import android.net.Uri;
import android.os.HandlerThread;
import android.support.test.runner.AndroidJUnit4;
import android.test.suitebuilder.annotation.SmallTest;

import com.android.systemui.SysuiTestCase;
import com.android.systemui.plugins.PluginInstanceManager.ClassLoaderFactory;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

@SmallTest
@RunWith(AndroidJUnit4.class)
public class PluginInstanceManagerTest extends SysuiTestCase {

    // Static since the plugin needs to be generated by the PluginInstanceManager using newInstance.
    private static Plugin sMockPlugin;

    private HandlerThread mHandlerThread;
    private Context mContextWrapper;
    private PackageManager mMockPm;
    private PluginListener mMockListener;
    private PluginInstanceManager mPluginInstanceManager;

    @Before
    public void setup() throws Exception {
        mHandlerThread = new HandlerThread("test_thread");
        mHandlerThread.start();
        mContextWrapper = new MyContextWrapper(getContext());
        mMockPm = mock(PackageManager.class);
        mMockListener = mock(PluginListener.class);
        mPluginInstanceManager = new PluginInstanceManager(mContextWrapper, mMockPm, "myAction",
                mMockListener, true, mHandlerThread.getLooper(), 1, true,
                new TestClassLoaderFactory());
        sMockPlugin = mock(Plugin.class);
        when(sMockPlugin.getVersion()).thenReturn(1);
    }

    @After
    public void tearDown() throws Exception {
        mHandlerThread.quit();
        sMockPlugin = null;
    }

    @Test
    public void testNoPlugins() {
        when(mMockPm.queryIntentServices(Mockito.any(), Mockito.anyInt())).thenReturn(
                Collections.emptyList());
        mPluginInstanceManager.startListening();

        waitForIdleSync(mPluginInstanceManager.mPluginHandler);
        waitForIdleSync(mPluginInstanceManager.mMainHandler);

        verify(mMockListener, Mockito.never()).onPluginConnected(
                ArgumentCaptor.forClass(Plugin.class).capture());
    }

    @Test
    public void testPluginCreate() {
        createPlugin();

        // Verify startup lifecycle
        verify(sMockPlugin).onCreate(ArgumentCaptor.forClass(Context.class).capture(),
                ArgumentCaptor.forClass(Context.class).capture());
        verify(mMockListener).onPluginConnected(ArgumentCaptor.forClass(Plugin.class).capture());
    }

    @Test
    public void testPluginDestroy() {
        createPlugin(); // Get into valid created state.

        mPluginInstanceManager.stopListening();

        waitForIdleSync(mPluginInstanceManager.mPluginHandler);
        waitForIdleSync(mPluginInstanceManager.mMainHandler);

        // Verify shutdown lifecycle
        verify(mMockListener).onPluginDisconnected(ArgumentCaptor.forClass(Plugin.class).capture());
        verify(sMockPlugin).onDestroy();
    }

    @Test
    public void testIncorrectVersion() {
        setupFakePmQuery();
        when(sMockPlugin.getVersion()).thenReturn(2);

        mPluginInstanceManager.startListening();

        waitForIdleSync(mPluginInstanceManager.mPluginHandler);
        waitForIdleSync(mPluginInstanceManager.mMainHandler);

        // Plugin shouldn't be connected because it is the wrong version.
        verify(mMockListener, Mockito.never()).onPluginConnected(
                ArgumentCaptor.forClass(Plugin.class).capture());
    }

    @Test
    public void testReloadOnChange() {
        createPlugin(); // Get into valid created state.

        // Send a package changed broadcast.
        Intent i = new Intent(Intent.ACTION_PACKAGE_CHANGED,
                Uri.fromParts("package", "com.android.systemui", null));
        mPluginInstanceManager.onReceive(mContextWrapper, i);

        waitForIdleSync(mPluginInstanceManager.mPluginHandler);
        waitForIdleSync(mPluginInstanceManager.mMainHandler);

        // Verify the old one was destroyed.
        verify(mMockListener).onPluginDisconnected(ArgumentCaptor.forClass(Plugin.class).capture());
        verify(sMockPlugin).onDestroy();
        // Also verify we got a second onCreate.
        verify(sMockPlugin, Mockito.times(2)).onCreate(
                ArgumentCaptor.forClass(Context.class).capture(),
                ArgumentCaptor.forClass(Context.class).capture());
        verify(mMockListener, Mockito.times(2)).onPluginConnected(
                ArgumentCaptor.forClass(Plugin.class).capture());
    }

    @Test
    public void testNonDebuggable() {
        // Create a version that thinks the build is not debuggable.
        mPluginInstanceManager = new PluginInstanceManager(mContextWrapper, mMockPm, "myAction",
                mMockListener, true, mHandlerThread.getLooper(), 1, false,
                new TestClassLoaderFactory());
        setupFakePmQuery();

        mPluginInstanceManager.startListening();

        waitForIdleSync(mPluginInstanceManager.mPluginHandler);
        waitForIdleSync(mPluginInstanceManager.mMainHandler);;

        // Non-debuggable build should receive no plugins.
        verify(mMockListener, Mockito.never()).onPluginConnected(
                ArgumentCaptor.forClass(Plugin.class).capture());
    }

    @Test
    public void testCheckAndDisable() {
        createPlugin(); // Get into valid created state.

        // Start with an unrelated class.
        boolean result = mPluginInstanceManager.checkAndDisable(Activity.class.getName());
        assertFalse(result);
        verify(mMockPm, Mockito.never()).setComponentEnabledSetting(
                ArgumentCaptor.forClass(ComponentName.class).capture(),
                ArgumentCaptor.forClass(int.class).capture(),
                ArgumentCaptor.forClass(int.class).capture());

        // Now hand it a real class and make sure it disables the plugin.
        result = mPluginInstanceManager.checkAndDisable(TestPlugin.class.getName());
        assertTrue(result);
        verify(mMockPm).setComponentEnabledSetting(
                ArgumentCaptor.forClass(ComponentName.class).capture(),
                ArgumentCaptor.forClass(int.class).capture(),
                ArgumentCaptor.forClass(int.class).capture());
    }

    @Test
    public void testDisableAll() {
        createPlugin(); // Get into valid created state.

        mPluginInstanceManager.disableAll();

        verify(mMockPm).setComponentEnabledSetting(
                ArgumentCaptor.forClass(ComponentName.class).capture(),
                ArgumentCaptor.forClass(int.class).capture(),
                ArgumentCaptor.forClass(int.class).capture());
    }

    private void setupFakePmQuery() {
        List<ResolveInfo> list = new ArrayList<>();
        ResolveInfo info = new ResolveInfo();
        info.serviceInfo = new ServiceInfo();
        info.serviceInfo.packageName = "com.android.systemui";
        info.serviceInfo.name = TestPlugin.class.getName();
        list.add(info);
        when(mMockPm.queryIntentServices(Mockito.any(), Mockito.anyInt())).thenReturn(list);

        when(mMockPm.checkPermission(Mockito.anyString(), Mockito.anyString())).thenReturn(
                PackageManager.PERMISSION_GRANTED);

        try {
            ApplicationInfo appInfo = getContext().getApplicationInfo();
            when(mMockPm.getApplicationInfo(Mockito.anyString(), Mockito.anyInt())).thenReturn(
                    appInfo);
        } catch (NameNotFoundException e) {
            // Shouldn't be possible, but if it is, we want to fail.
            throw new RuntimeException(e);
        }
    }

    private void createPlugin() {
        setupFakePmQuery();

        mPluginInstanceManager.startListening();

        waitForIdleSync(mPluginInstanceManager.mPluginHandler);
        waitForIdleSync(mPluginInstanceManager.mMainHandler);
    }

    private static class TestClassLoaderFactory extends ClassLoaderFactory {
        @Override
        public ClassLoader createClassLoader(String path, ClassLoader base) {
            return base;
        }
    }

    // Real context with no registering/unregistering of receivers.
    private static class MyContextWrapper extends ContextWrapper {
        public MyContextWrapper(Context base) {
            super(base);
        }

        @Override
        public Intent registerReceiver(BroadcastReceiver receiver, IntentFilter filter) {
            return null;
        }

        @Override
        public void unregisterReceiver(BroadcastReceiver receiver) {
        }
    }

    public static class TestPlugin implements Plugin {
        @Override
        public int getVersion() {
            return sMockPlugin.getVersion();
        }

        @Override
        public void onCreate(Context sysuiContext, Context pluginContext) {
            sMockPlugin.onCreate(sysuiContext, pluginContext);
        }

        @Override
        public void onDestroy() {
            sMockPlugin.onDestroy();
        }
    }
}
